#####
###  PHParser - A Gateway to PHP
##   Copyright (C) 2003	Wei He (hewei@ied.org.cn)
#    Copyright (C) 2002	Maik Jablonski (maik.jablonski@uni-bielefeld.de)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
##   along with this program; if not, write to the Free Software
###  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#####

import sys, os, string

#
# Path to the PHP-CGI-Interpreter.
# Modify it to where your PHP CGI program is located. PHP CLI does not work.
#                             ^^^
if sys.platform == 'win32':
    PHPath = "c:\\php\\php.exe"
    from win32pipe import popen2
else:
    ENVPath = "env"
    PHPath = "/var/www/cgi-bin/php"
    from os import popen2

#
# Whether to expose HTTP authentication info to the PHP script.
#
#######################################################################
# Warning: Turning this on may pass your ZMI username/password to     #
#          the PHP script when the script is parsed.                  #
#          Do not turn this option on if you don't trust the script.  #
#######################################################################
#
EXPOSE_HTTP_AUTHORIZATION = 0

PHPEOL = "\r?\n"

from Globals import HTML, DTMLFile, InitializeClass
from OFS.content_types import guess_content_type
from OFS.DTMLMethod import decapitate
from OFS.DTMLDocument import DTMLDocument
from AccessControl import getSecurityManager
from ZPublisher.HTTPRequest import FileUpload
from urllib import quote, urlencode
import re, tempfile, copy

_marker=[]

#
# Take a snapshot of the environment on Win32 platform.
#
if sys.platform == 'win32':
    _envsave = copy.deepcopy(os.environ)
    _envkeys = _envsave.keys()

#
# The default content of a newly added PHParser object.
#
default_dd_html="""
<dtml-var standard_html_header>

<?php
	echo "<h2><dtml-var title_or_id> / {$ZOPE_VARS['BASE0']}$PHP_SELF</h2>";
	echo "<h2>PHParser extended global variables: </h2><pre>\$ZOPE_VARS=";
	print_r($ZOPE_VARS);
	echo "</pre>";
	phpinfo();
?>

<dtml-var standard_html_footer>
"""

addPHPForm = DTMLFile('dtml/addPHPForm', globals())

#
# This code is taken and slightly modified from OFS/DTMLDocument.py
#
def addPHParser(self, id, title='', file='', REQUEST=None, submit=None):
    """Add a PHParser object with the contents of file. If
    'file' is empty, default document text is used.
    """
    if type(file) is not type(''): file = file.read()
    if not file: file = default_dd_html
    id = str(id)
    title = str(title)
    ob = PHParser(file, __name__=id)
    ob.title = title

    id = self._setObject(id, ob)
    if REQUEST is not None:
        try: u = self.DestinationURL()
        except: u = REQUEST['URL1']
        if submit == " Add and Edit ": u="%s/%s" % (u,quote(id))
        REQUEST.RESPONSE.redirect(u + '/manage_main')
    return ''

#
# The class defination of PHParser
#
class PHParser(DTMLDocument):
    """ PHParser: It's a DTMLDocument which will be parsed by PHP after
        rendering the DTML """

    meta_type = 'PHParser'
    manage_options = DTMLDocument.manage_options
    
    #
    # Transform Python variables into PHP expression
    #
    def transformValue(self, value):

        if type(value) in (type(0), type(0.0)):
            #
            # number
            #
            result = str(value)

        elif type(value) in (type([]), type(())):
            #
            # list or turple
            #
            result = "array("
            if len(value) > 0:
                for v in value:
                    result += self.transformValue(v) + ","
                result = result[:-1]
            result += ")"

        elif type(value) == type({}):
            #
            # dict
            #
            result = "array("
            if len(value) > 0:
                for (k,v) in value.items():
                    result += self.transformValue(str(k)) + "=>" + self.transformValue(v) + ","
                result = result[:-1]
            result += ")"

        else:
            #
            # string and others
            #
            result = "'" + string.replace(str(value), "'", "\\'") + "'"

        return result

    #
    # Build a temporary PHP script file to be parsed and assign the name
    # to self.__script__ or set self.__result__ to skip parsing.
    # The retured is the CGI environment.
    #
    def __prepare__(self, src, REQUEST={}):
        varfix = ""

        """ Force REQUEST to generate all items. Any better idea? """
        REQUEST.text()

        """ Construct $ZOPE_VARS from REQUEST.other """
        zopevars = {}
        for item in REQUEST.other.keys():
            if not item in ('RESPONSE', 'PUBLISHED', 'PARENTS'):
                 zopevars.update({item: REQUEST.other[item]})
        varfix += "$ZOPE_VARS=" + self.transformValue(zopevars) + "; "

        """ Always expose $PHP_SELF """
        varfix += "$PHP_SELF = $_SERVER['PHP_SELF'] = $HTTP_SERVER_VARS['PHP_SELF']; "

        """ In case you want to prepend your own codes. """
        #varfix += "include('" + INSTANCE_HOME + "/php.inc'); "

        """ Create a temporary PHP script on disk """
        self.__script__ = tempfile.mktemp(".php")
        #print "PHPARSER: Creating temp file: " + self.__script__ 
        ftemp = open(self.__script__, "wb")
        ftemp.write('<?php ' + varfix + ' ?>')
        ftemp.write(src)
        ftemp.close()
   
        """ Return constructed CGI environment """
        return self.__buildenv__(REQUEST)

    #
    # Build an environment for calling PHP CGI.
    #
    def __buildenv__(self, REQUEST={}):
        cgienv = {}

        for (k,v) in REQUEST.environ.items():
            cgienv[k] = str(v)

        # todo: Virtual host support not yet tested
        cgienv['SCRIPT_NAME'] = '/' + self.absolute_url(1)

        """ These two are required by PHP CGI """
        cgienv['SCRIPT_FILENAME'] = self.__script__
        cgienv['PATH_TRANSLATED'] = self.__script__

        cgienv['REDIRECT_STATUS'] = '200'

        """ Apache generates these two """
        #cgienv['REDIRECT_SCRIPT_URL'] = cgienv['SCRIPT_NAME']
        #cgienv['REDIRECT_SCRIPT_URI'] = REQUEST['URL0']

        """ Emulate the apache-specific REQUEST_URI """
        cgienv['REQUEST_URI']=string.replace(REQUEST['URL0'], REQUEST['BASE0'], '')
        if (cgienv['PATH_INFO'][-1] == '/'):
           cgienv['REQUEST_URI'] += '/'
        if (cgienv.has_key('QUERY_STRING')):
           cgienv['REQUEST_URI'] += '?' + cgienv['QUERY_STRING']
        
        """ A single PHParser script shouldn't return PATH_INFO """
        del cgienv['PATH_INFO']

        """ Expose the HTTP authentication info """
        if EXPOSE_HTTP_AUTHORIZATION and REQUEST._auth:
            cgienv['HTTP_AUTHORIZATION'] = REQUEST._auth

        """ Reconstruct the form data (fields and files). """
        #
        # Thanks to 
        # http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/146306
        #
        if string.upper(cgienv['REQUEST_METHOD']) == 'POST':
            multipart = re.search("^multipart/form-data; boundary=(.*)", cgienv['CONTENT_TYPE'])
            if (multipart):
               #
               # form data is posted as multipart/form-data
               #
               BOUNDARY=multipart.groups()[0]
               CRLF = '\r\n'
               L = []
               for (k,v) in REQUEST.form.items():
                  L.append('--' + BOUNDARY)
                  if isinstance(v, FileUpload):
                     #
                     # handle file uploads
                     #
                     L.append('Content-Disposition: form-data; name="%s"; filename="%s"' % (k, v.filename))
                     L.append('Content-Type: %s' % v.headers['content-type'])
                     content=v.read()
                  else:
                     #
                     # handle input fields
                     #
                     L.append('Content-Disposition: form-data; name="%s"' % k)
                     content=v
                  L.append('')
                  L.append(content)
               L.append('--' + BOUNDARY + '--')
               L.append('')
               self.__formdata__ = CRLF.join(L)
            else:
               #
               # form data is posted as application/x-url-encoded
               #
               self.__formdata__ = urlencode(REQUEST.form, 1)

            cgienv['CONTENT_LENGTH'] = str(len(self.__formdata__))
        else:
            self.__formdata__ = None

        return cgienv

    #
    # This is the function where the document is piped through PHP
    #
    def __phparse__(self, src, REQUEST={}):

        """ If self.__script__ is set after __prepare__,
            the script is going to be deleted after parsing. """

        self.__script__ = ''

        """ If self.__result__ is set after __prepare__, parsing
            will be skipped and the result will be returned as-is. """
 
        self.__result__ = None

        env = self.__prepare__(src, REQUEST)
        
        if self.__result__ != None:
            #print "PHPARSER: Skip PHParsing."
            return (self.__result__)
            
        #print "PHPARSER: PHParsing..."
        #print "**CGI ENV:"
        #for (k,v) in env.items():
        #    print '  ' + k + '=' + v

        if sys.platform == 'win32':
            # Win32 does not have a 'env' program to provide
            # a modified environment. Here we modify the main environment
            # and later restore it to the snapshot copy.
            for k in os.environ.keys():
                """ These variables seem to be important. """
                if string.lower(k) not in ('path', 'comspec', 'systemroot', 'windir', 'temp', 'tmp'):
                    os.environ[k] = ''
                    del os.environ[k]
            os.environ.update(env)
            cmd = PHPath
        else:
            # Posix platforms take advancanges of the 'env' program
            # to run PHP CGI in a modified environment.
            cmd = ENVPath + ' -i '
            for (k,v) in env.items():
                cmd += k + '="' + v + '" '
            cmd += PHPath

        
        """ Ladies and gentlemen... and now... PHParsing... """
        (stin,stout) = popen2(cmd, 'b')

        """ form data and files are to be written to stdin, while the
            PHP script is a temporary file whose name is sepcified by 
            the CGI environment 'SCRIPT_FILENAME'. """
        if self.__formdata__:
            stin.write(self.__formdata__)
            stin.flush()
        stin.close()

        """ Get the raw result from PHP CGI. """
        raw = stout.read()
        stout.close()
        
   
        """ Repair the environment on win32 platform. """
        if sys.platform == 'win32':
            for k in os.environ.keys():
                if k in _envkeys:
                    os.environ[k] = _envsave[k]
                else:
                    os.environ[k] = ''
                    del os.environ[k]
                    
        """ Delete the temporary PHP script. """
        if self.__script__ != '':
            os.unlink(self.__script__)

        """ Now decode the Header and Body of the result """
        phpresult = re.search("(?s)(?P<header>^.*?)%s(?P<body>.*)"%(PHPEOL*2), raw)
        if phpresult:
            """ Trust the headers returned by PHP. """
            (phpHeader,phpBody) = phpresult.groups()
            #print "PHPARSER: PHP headers: " + phpHeader
            for line in re.split(PHPEOL, phpHeader):
                if ':' in line:
                    (name,value) = line.split(':',1)
                    if name.lower() == 'location':
                        REQUEST.RESPONSE.redirect(value.strip())
                    else:
                        REQUEST.RESPONSE.setHeader(name,value.strip())
                        #if name.lower() == 'content-type':
                        #    print "PHPARSER: Content-Type: " + value.strip()
        else:
            """ Generate a friendly warning message rather than 0-byte. """
            REQUEST.RESPONSE.setHeader("Content-type", "text/plain")
            phpBody = "'" + PHPath + "' may not be a PHP-CGI program. The result from the program is"
            if len(raw):
                phpBody += ":\n" + raw
            else:
                phpBody += " empty(0 byte)."

        return(phpBody)

    #
    # This is the original code from OFS/DTMLDocument.py
    # Only a call to __phparse__ is added (see below)
    #
    # It has not been modified since the initial release of 0.20413 
    # by Maik Jablonski
    #
    def __call__(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """Render the document given a client object, REQUEST mapping,
        Response, and key word arguments."""


        if not self._cache_namespace_keys:
            data = self.ZCacheable_get(default=_marker)
            if data is not _marker:
                # Return cached results.
                return data


        kw['document_id']=self.getId()
        kw['document_title']=self.title
        if hasattr(self, 'aq_explicit'):
            bself=self.aq_explicit
        else: bself=self

        security=getSecurityManager()
        security.addContext(self)

        try:
            if client is None:
                # Called as subtemplate, so don't need error propigation!
                r = apply(HTML.__call__, (self, bself, REQUEST), kw)
                r = self.__phparse__(r, bself.REQUEST)
                if RESPONSE is None: result = r
                else: result = decapitate(r, RESPONSE)
                if not self._cache_namespace_keys:
                    self.ZCacheable_set(result)
                return result

            r = apply(HTML.__call__, (self, (client, bself), REQUEST), kw)
            r = self.__phparse__(r, bself.REQUEST)

            if type(r) is not type('') or RESPONSE is None:
                if not self._cache_namespace_keys:
                    self.ZCacheable_set(r)
                return r

        finally: security.removeContext(self)

        have_key=RESPONSE.headers.has_key
        if not (have_key('content-type') or have_key('Content-Type')):
            if self.__dict__.has_key('content_type'):
                c=self.content_type
            else:
                c, e=guess_content_type(self.__name__, r)
            RESPONSE.setHeader('Content-Type', c)
        result = decapitate(r, RESPONSE)
        if not self._cache_namespace_keys:
            self.ZCacheable_set(result)
        return result

InitializeClass(PHParser)

